;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Emacs CPerl and Perl Mode Additional Features For working with Perl Code
;;

;; add the following to your .emacs file to use these utilities:
;;
;; (setq load-path (cons (getenv "HOME") "/projects/hmsdev2/etc/" load-path))
;;

;; Accidentaly exiting emacs when you didn't mean it is _annoying_ To
;; use this, put it in your .emacs
;;
;; (defun hms-save-buffers-kill-emacs (answer)
;; (interactive "sReally Quit? (yes/no): ")
;; (if (string-equal "yes" answer)
;; (save-buffers-kill-emacs)))
;; (global-set-key "\C-x\C-c" 'hms-save-buffers-kill-emacs)

;; These utilities make use of some CommonLisp features
(require 'cl)

;; Other modules that we rely on
(load "dabbrev")
(load "completion")
(initialize-completions)

; use the following to override the keybinding for the CVS checkin...
; (global-set-key "\C-xvv" 'hms-vc-checkin-wrapper)


(defun hms-perl-find-log4perl-config-file ()
  "Finds the log4perl config file by looking for the env var first, then defaulting to ~/.log4perl.conf"
  (cond ((file-exists-p (getenv *hms-perl-log4perl-config-env-var*))
         (getenv *hms-perl-log4perl-config-env-var*))
        ((file-exists-p (expand-file-name "~/.log4perl.conf"))
         (expand-file-name "~/.log4perl.conf"))
        (t
         nil)))

;;;;
;;;; Parameter Declaration
;;;;

;; set this to false to turn of the addition of pod when you create
;; new modules or add new methods
;; 
;;   (setq *hms-perl-include-pod* nil)
;; 
(defvar *hms-perl-include-pod* t)

; user information
(defvar *hms-short-username* (getenv "LOGNAME"))
(defvar *hms-user-full-name* user-full-name)
(defvar *hms-email-address* user-mail-address)

(defvar *hms-perl-source-path-prefix* "/lib/")
(defvar *hms-perl-vendor-package-prefix* "HMS")
(defvar *hms-perl-standard-base-class* (concat *hms-perl-vendor-package-prefix* "::Base"))
(defvar *hms-perl-keybinding-prefix* "\C-xp")

(defvar *hms-perl-emacs-utils-file-path* (concat (getenv "HMS_ROOT") "/etc/emacs-utils.el"))

(defvar *hms-perl-log4perl-config-env-var* "LOG4PERL_CONFIG")
(defvar *hms-perl-log4perl-config-file* (hms-perl-find-log4perl-config-file))

(defvar *hms-perl-regex-perlvariables* "[$%@][a-zA-Z0-9_]\+")
(defvar *hms-perl-regex-perlpackages* "[A-Z][a-zA-Z0-9]\+\\(::[A-Z][A-Za-z0-9_]\*\\)\+")
(defvar *hms-perl-regex-constant-decls* "use constant [A-Z][A-Z_0-9]\+ \+=>")

(defvar *hms-vc-email-address* user-mail-address)
(defvar *hms-vc-last-used-bug-id* nil)

(defvar *hms-perl-test-unit-test-runner-cmd* "TestRunner.pl")
(defvar *hms-perl-test-suite-runner-cmd* "runUnitTests.pl")
(defvar *hms-perl-test-buffer-name* "*Shell Command Output*")

(defvar *hms-perl-unit-test-cmdline-func* 'hms-perl-make-test-runner-cmdline
  "The name of the function which generates the unit test runner shell
command.  You can override this by setting your own function.  The
function should expect a single argument, the name of the class to be
tesed as a Perl package name.

(defun my-test-runner (package)
  (interactive)
  (format \"ssh user@host 'bash ~/my-runner.sh %s'\" package))
(setf *hms-perl-unit-test-cmdline-func* 'my-test-runner)

")

(defvar *hms-perl-test-suite-cmdline-func* 'hms-perl-make-test-suite-runner-cmdline
  "The name of the function which generates the test suite runner
shell command.  You can override this by setting your own function.
The function will not be passed any arguments.

(defun my-test-suite-runner (package)
  (interactive)
  (format \"ssh user@host 'bash ~/my-suite-runner.sh %s'\" package))
(setf *hms-perl-test-suite-cmdline-func* 'my-test-suite-runner)

")

(defvar *hms-perl-keybinding-specs* '())

(defvar *hms-perl-reworked-block-comment-start* " # ==> { start of reworked block")
(defvar *hms-perl-reworked-block-comment-mid* " # }{ midpoint of reworked block")
(defvar *hms-perl-reworked-block-comment-end* " # } <== end of reworked block")

(defvar *hms-perl-tags-file-stack* (list))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Utility Macros used by other functions.
;;
(defmacro hms-perl-make-setting-prompter (fname sym prompt default)
"This macro creates a function that will prompt the user for a value
if the symbol has no value."
`(defun ,fname ()
   (concat "This function will return the value of " (symbol-name ',sym) ". If " (symbol-name ',sym) " is null, it will prompt the user for a value and store it.")
   (interactive)
   (when (not ,sym)
     (setq ,sym (read-string ,prompt ,default)))
   ,sym))


(defmacro hms-shift (lat)
  "Destructive shift off of the right hand side of the list of atoms.
As opposed to pop which works on the left."
  `(let ((rev (reverse ,lat)))
     (setq ,lat (reverse (cdr rev)))
     (car rev)))

(defmacro hms-defun-with-keybinding (sym keysuffix &rest body)
  "This is a wrapper around defun that additionaly registers a keybinding for 
the symbol."
  (push (list keysuffix sym) *hms-perl-keybinding-specs*)
  `(defun ,sym ,@body))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Additional VC Hooks and Wrappers
;;
(defun hms-vc-checkin-wrapper ()
  "Wrapper around vc-next-action to insert a few lines into the *VC-log* 
window to prepopulate it."
  (interactive)
  (vc-next-action nil)
  (let ((vc-buffer (get-buffer "*VC-log*")))
    (if vc-buffer
        (progn
          (switch-to-buffer vc-buffer)
          (insert (concat "Jira ID: " (hms-perl-vc-get-last-used-bug-id) "\n"))
                          "\n\n"))))

(hms-perl-make-setting-prompter hms-perl-vc-get-email-address 
                                *hms-vc-email-address* "Email Address: " "@hmsonline.com<http://hmsonline.com>")

(hms-perl-make-setting-prompter hms-perl-vc-get-last-used-bug-id 
                                *hms-vc-last-used-bug-id* "Bug ID: " "")



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Emacs Mode Hooks and Configuration
;;;;

;; use spaces instead of tabs
(setq-default indent-tabs-mode nil)

;; use a tabstop of 2 when cc-mode auto-indents
(setq c-default-style "user"
      c-basic-offset 4)

(setq tags-table-list
      `( ,(concat (getenv "HMS_ROOT") "/TAGS" )))


;;
;; When we enter CPerl mode or Perl mode, execute our hook
;;
(add-hook 'cperl-mode-hook 'hms-cperl)
(defun hms-cperl ()
  "Configure our editor settings, including indentation and keybindings."
  (interactive)
  (setq cperl-indent-level 2)
  (setq cperl-continued-brace-offset 0)
  (setq cperl-merge-trailing-else nil)
  (hms-cperl-apply-keybindings))

(add-hook 'perl-mode-hook 'hms-perl)
(defun hms-perl ()
  (interactive)
  (setq perl-indent-level 2)
  (setq perl-continued-brace-offset 0)
  (setq perl-merge-trailing-else nil)
  (hms-cperl-apply-keybindings))

(defun hms-perl-show-keybinding-list ()
  (interactive)
  (message "all keys: %s" *hms-perl-keybinding-specs*)
  (let ((msg ""))
    (loop for pair in *hms-perl-keybinding-specs*
          do
          (setq msg (concat msg (format "Key: %s => %s\n" (first pair) (second pair)))))
    (message msg)))

;;
;; Set up keybindings for CPerl mode
;;
(defun hms-cperl-apply-keybindings ()

  ;; All keybindings are prefixed with: CTRL+X p
  (mapcar
   (lambda (pair)
     (local-set-key (concat *hms-perl-keybinding-prefix* (first pair)) (second pair)))
   '(("c" hms-perl-compile-buffer)                        ; c - 'c'ompile current file
     ("C" hms-perl-export-all-constants)                  ; C - export all 'C'onstants
     ("\C-c" hms-perl-insert-constant)                    ; ^C - add a constant
     ("d" hms-perl-doc-method)                            ; d - pod api 'd'ocument current method
     ("D" hms-perl-perldoc-buffer)                        ; D - view POD for an open buffer/file
     ("\C-d" hms-perl-insert-module-header-pod)           ; D - insert module header pod
     ("e" hms-perl-insert-exception)                      ; e - insert a 'throw' statement,'e'xception
     ("E" hms-perl-insert-try-catch)                      ; E - insert a 'throw' statement,'E'xception
     ("f" hms-perl-insert-foreach)                        ; f - insert a 'f'oreach template
     ("i" hms-perl-insert-if)                             ; i - insert an 'i'f template
     ("I" hms-perl-push-inc)                              ; I - push new path onto @'I'NC
     ("\C-i" hms-perl-insert-identity-constant)           ; ^i - Add 'i'dentity constant to MakeIdentityConstants list
     ("j" hms-perl-insert-join)                           ; j - insert a join statement
     ("J" hms-perl-justify-line-at-token)                 ; J - internaly 'j'ustify a line's contents
     ("\C-j" join-line)                                   ; ^j - 'j'oin line
     ("l" hms-perl-insert-log-statement)                  ; l - insert l'og method
     ("L" hms-perl-adjust-logging-for-package)            ; L - adjust logging for package
     ("\C-l" hms-perl-open-log4perl-config)               ; ^l - open log4perl configuration file
     ("\M-l" hms-perl-set-new-log4perl-config)            ; ALT-l - set new log4perl config file
     ("m" hms-perl-insert-method)                         ; m - insert new 'm'ethod
     ("M" hms-perl-new-module)                            ; M - create new 'M'odule in current buffer
     ("n" hms-perl-find-next-error)                       ; n - 'n'ext error in the unit test output buffer
     ("\C-o" hms-perl-open-new-line-above)                ; o - open a new line, kinda like C-o, but also jump to the beginning
     ("p" hms-perl-insert-basic-perl-script)              ; p - run hms-perl-insert-basic-'p'erl-script
     ("P" hms-perl-refactor-push-method-to-parent)        ; P - 'P'ush method to parent
     ("r" hms-perl-prev-sub)                              ; r - previous sub declaration ('r'everse search)
     ("R" hms-perl-refactor-selection-as-method)          ; R - 'R'efactor selection as method
     ("\C-r" hms-perl-reload-emacs-utils)                 ; R - 'R'eload the emacs utilities...
     ("s" hms-perl-follow-stack-trace)                    ; s - follow through a 's'tack trace
     ("S" hms-perl-follow-unit-test-stack-trace)          ; S - follow through a 'S'tack trace
     ("T" hms-perl-switch-to-unit-test)                   ; T - module unit 'T'est
     ("t" hms-perl-add-unit-test-for-method)              ; t - method unit 't'est
     ("\C-t" hms-perl-run-unit-test)                      ; ^t - run the unit test for current module
     ("\M-t" hms-perl-print-test-runner-cmdline)          ; ALT-t - show tester cmdline
     ("u" hms-perl-insert-unless)                         ; u - insert an 'u'nless template
     ("U" hms-perl-fix-use-modules)                       ; U - add 'U'se statements for all packages
     ("v" hms-perl-new-member)                            ; v - new member 'v'ariable
     ("V" hms-perl-refactor-selection-as-local-variable)  ; V - refactor selection as 'V'ariable
     ("w" hms-perl-insert-while)                          ; w - insert a 'w'hile template
     ("y" hms-perl-copy-and-comment-region)               ; y - Copy and comment the selected region
     ("." hms-perl-open-buffer-for-package)               ; . - like Alt-. (like Tags)
     ("*" hms-perl-pop-tags-stack)                        ; * - like Alt-* (like Tags)
     ("}" hms-perl-add-else-clause)                       ; '}' add closing else clause
     ))
  (message "keybindings applied"))


;;;;
;;;; Utility Function Definitions
;;;;

(hms-perl-make-setting-prompter 
 hms-perl-get-email-address 
 *hms-email-address* "Email Address: " "@healthmarketscience.com")

(hms-perl-make-setting-prompter 
 hms-perl-get-user-full-name 
 *hms-user-full-name* "Your Full Name: " "")

(defun hms-perl-get-name-of-current-sub ()
  (interactive)
  (save-excursion
        (let ((start) 
              (sub-name))
          (hms-perl-prev-sub)
          (forward-word 1)
          (forward-char 1)
          (setq start (point))
          (end-of-line)
          (search-backward-regexp " \\|{\\\n")
          (setq sub-name (buffer-substring start (point)))
        sub-name)))

(defun hms-perl-reload-emacs-utils ()
  "Re-loads the emacs-utils.el file (based on the location of emacs-utils.el) 
and re-inits cperl-mode"
  (interactive)
  (load *hms-perl-emacs-utils-file-path*)
  (cperl-mode))

; from mct
(defun goto-percent (pct)
  "Computes the character position of the given percentage of the current 
buffer and places the cursor at that position."
  (interactive "nGoto percent: ")
  (let* ((size (point-max))
         (charpos (/ (* size pct) 100)))
    (goto-char charpos)
    (message "Moved to charpos: %d/5d." charpos size)))

; from mct
;(global-set-key "%" 'match-paren)
(defun match-paren (arg)
  "Go to the matching paren if on an open paren; otherwise insert %."
  (interactive "p")
  (cond ((looking-at "\\s\(")
         (forward-list 1)
         (backward-char 1))
        ((looking-at "\\s\)")
         (forward-char 1)
         (backward-list 1))
        (t (self-insert-command (or arg 1)))))

(defun hms-perl-next-sub ()
  "Attempts to move the pointer to the next subroutine (simple quick regex 
search)"
  (interactive)
  (search-forward-regexp "^\\s-*sub "))

(defun hms-perl-prev-sub ()
  "Attempts to move the pointer to the previous subroutine (simple quick regex 
search)"
  (interactive)
  (search-backward-regexp "^\\s-*sub "))

(defun hms-perl-doc-method ()
  "Attempts to POD a method"
  (interactive)
  ;; find the sub name
  (let ((sub-name) (args) (args-no-self) (pod-start-point))
    (search-backward-regexp "^sub")
    (setq pod-start-point (point))
    (forward-word 1)
    (forward-char 1)
    (let ((beg (point)))
      (search-forward-regexp "\\( \\|{\\|$\\)")
      (backward-char 1)
      (setq sub-name (buffer-substring beg (point))))
    (condition-case nil
        (let ((beg))
          (search-forward-regexp "my(")
          (setq beg (point))
          (search-forward-regexp ")")
          (backward-char 1)
          (setq args (buffer-substring beg (point))))
      (error nil))
    (goto-char pod-start-point)
    (beginning-of-line)
    (if args
        (setq args-no-self (substring args (+ (hms-string-find args "," 0) 1))))
    (when (not args-no-self) 
      (setq args-no-self args))
    (insert "\n")
    (insert
     (hms-string-join "\n" (list
                            (concat "=head2 " sub-name)
                            (if args
                                (concat "\n  @param:   "
                                        (hms-string-join " - OBJECT::TYPE\n  @param:   "  (split-string args-no-self ",")) " - OBJECT::TYPE" ))
                            "  @returns: true/false"
                            "  @throws:  HMS::Exception upon failure"
                            ""
                            (concat "  my $rv = $self->" sub-name "(" args-no-self ");")
                            ""
                            (concat "Describe " sub-name " here.")
                            ""
                            "=cut"
                            ""
                            "")))
    (previous-line 4)
    (beginning-of-line)
    (recenter)
    (message "Document the method if you don't mind.")))

(defun hms-perl-word-under-cursor ()
  "Returns the 'word' at the point."
  (interactive)
  (save-excursion
    (backward-word 1)
    (let ((beg (point)))
      (forward-word 1)
      (buffer-substring beg (point)))))

(defun hms-perl-position-pointer-at-method-insertion-point ()
  "Moves the pointer to the spot in the current buffer just before the
end of the module (the 1;)"
  (interactive)
  (goto-char (point-max))
  (search-backward "1;")
  (backward-char 1))

(defun hms-perl-insert-method (method-name arguments)
  "Inserts a new method into the buffer at the current locaiton of the 
poitner."
  (interactive (list (read-string "Method Name: " (hms-perl-word-under-cursor))
                     (read-string "Arguments in addition to $self: " 
                                  (hms-perl-find-next-arg-list))))
  (hms-perl-position-pointer-at-method-insertion-point)
  (hms-perl-insert-generated-method method-name arguments))

(defun hms-perl-insert-generated-method (method-name arguments)
  (interactive (list (read-string "Method Name: " (hms-perl-word-under-cursor))
                     (read-string "Arguments in addition to $self: " 
                                  (hms-perl-find-next-arg-list))))
  (let ((line-count 2) (args-no-self arguments))
    (if (not (= 0 (length arguments)))
        (setq arguments (concat "$self," arguments))
      (setq arguments "$self"))
    (hms-perl-insert-indented "sub " method-name " {\n  my(" arguments ") = @_;\n")
    (hms-perl-insert-indented "  HMS::Base->throw(HMS::UnimplementedMethodException->new(join('',\"implement this\")));\n")
    (hms-perl-insert-indented "}\n")
    (previous-line line-count)
    (end-of-line)
    (recenter)
    (if *hms-perl-include-pod*
        (hms-perl-doc-method))
    (message "Document the method if you don't mind.")))

(defun hms-perl-gen-module-header-pod (module-name)
  "Generate and return basic module documentation (name, synopsis, etc.)."
  (interactive (list (read-string "Module Name: " (hms-perl-get-package-name-for-current-buffer))))
  (hms-string-join "\n" (list
                         "=head1 NAME"
                         ""
                         (concat module-name " - This Class is For...")
                         ""
                         "=head1 SYNOPSIS"
                         ""
                         "  use strict;"
                         "  use warnings;"
                         (concat "  use " module-name ";")
                         ""
                         (concat "  my $instance = " module-name "->new(some set of parameters);")
                         ""
                         "=head1 DESCRIPTION"
                         ""
                         (concat "This module has to be documented by " (hms-perl-get-user-full-name) " <" (hms-perl-get-email-address) ">.")
                         ""
                         "=head1 CONSTANTS"
                         ""
                         "=head1 MEMBERS"
                         ""
                         "=over"
                         ""
                         "=back"
                         ""
                         "=head1 EXPORTS"
                         ""
                         "=head1 API REFERENCE"
                         ""
                         "=cut"
                         "")))

(defun hms-perl-insert-module-header-pod (module-name)
  "Inserts the returned string from hms-perl-gen-module-header-pod."
  (interactive (list (read-string "Module Name: " (hms-perl-get-package-name-for-current-buffer))))
  (insert (hms-perl-gen-module-header-pod module-name)))
  
(defun hms-perl-gen-module-footer-pod (module-name)
  "Generate and return a basic POD footer for the given module."
  (interactive (list (read-string "Module Name: " (hms-perl-get-package-name-for-current-buffer))))
  (hms-string-join "\n" (list
                         ""
                         "=head1 ENVIRONMENT"
                         ""
                         "=head1 KNOWN BUGS"
                         ""
                         "=head1 SEE ALSO"
                         ""
                         "=head1 AUTHORS"
                         ""
                         (concat (hms-perl-get-user-full-name) " <" (hms-perl-get-email-address) ">")
                         ""
                         "=cut"
                         "")))
  
(defun hms-perl-insert-module-footer-pod (module-name)
  "Inserts the returned string from hms-perl-gen-module-footer-pod."
  (interactive (list (read-string "Module Name: " (hms-perl-get-package-name-for-current-buffer))))
  (insert (hms-perl-gen-module-footer-pod module-name)))

(defun hms-perl-insert-basic-perl-script ()
  "Boiler plat for a basic new Perl script."
  (interactive)

  (hms-perl-insert-indented "package _TMP;\n")
  (hms-perl-insert-indented "use strict;\n")
  (hms-perl-insert-indented "use warnings;\n")
  (hms-perl-insert-indented "our $CVS_ID = '$Id: emacs-utils.el,v 1.60 2008/03/11 19:58:48 jcrean Exp $';\n")
  (hms-perl-insert-indented "our $VERSION = ( qw$Revision: 1.60 $ )[1];\n")
  (hms-perl-insert-indented "use base qw(HMS::Util::App);\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "use HMS::Base;\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "my $Log = HMS::Base->get_log_context;\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "HMS::Base->MakeGSS(qw( ));\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "sub autoCmdLineArgs {\n")
  (hms-perl-insert-indented "my($self) = @_;\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "return ($self->SUPER::cmdLineArgs,\n")
  (hms-perl-insert-indented "qw(\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "));\n")
  (hms-perl-insert-indented "}\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "sub run {\n")
  (hms-perl-insert-indented "my($self) = @_;\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "}\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "1;\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "_TMP->main;\n")
  (end-of-buffer)
  (previous-line 6)
  (insert "  "))

(defun hms-perl-new-module (module-name)
  "Boiler plate for a new Perl module."
  (interactive (list (read-string "Module Name: " (hms-compute-package-from-pwd))))
  (insert "# -*- mode: CPerl -*-\n")
  (insert "package " module-name ";\n")
  (insert "use strict;\n")
  (insert "use warnings;\n")
  (insert "our $CVS_ID = '$Id")           (insert "$';\n")  ; using 2 inserts in this manner avoids
  (insert "our $VERSION = ( qw$Revision") (insert "$ )[1];\n") ; CVS tag expansion for these strings
  (insert "use base qw(" *hms-perl-standard-base-class* ");\n")
  (insert "\n")
  (insert "use " *hms-perl-standard-base-class* ";\n")
  (insert "\n")
  (insert "my $Log = " *hms-perl-standard-base-class* "->get_log_context;\n")
  (insert "\n")
  (insert *hms-perl-standard-base-class* "->MakeGSS(qw( ));\n")
  (insert "\n")
  (insert "\n")
  (if *hms-perl-include-pod* (insert (hms-perl-gen-module-header-pod module-name)))
  (insert "# start your methods here\n")
  (insert "\n")
  (insert "\n")
  (insert "1;\n")
  (if *hms-perl-include-pod* (insert (hms-perl-gen-module-footer-pod module-name)))
  (previous-line 2)
  (recenter)
  (message "Don't forget to document the package and the API."))

(defun hms-perl-package-to-directory (package)
  (message "to directory: " package)
  (let ((parts (split-string package "::")))
    (hms-shift parts)
    (hms-string-join "/" parts)))

(defun hms-perl-compute-test-file-for-current-buffer ()
  (let* ((test-directory "Test")
         (test-package (hms-perl-make-test-package-name 
                        (hms-perl-get-package-name-for-current-buffer))))
    (concat test-directory "/" (car (last (split-string test-package "::"))) ".pm" )))

(defun hms-perl-buffer-is-unit-test ()
  "Returns true/nil if the buffer is a unit test module (package name contains ::Test::"
  (interactive)
  (not (= -1 (hms-string-find (hms-perl-get-package-name-for-current-buffer) "::Test::" 0))))

(defun hms-perl-switch-to-unit-test ()
  "Switches between the test and the packae to be tested."
  (interactive)
  (message "is test? %s" (hms-perl-buffer-is-unit-test))
  (cond ((hms-perl-buffer-is-unit-test)
         (let ((file-name (buffer-file-name)))
           (setf file-name (replace-regexp-in-string "/Test/\\(.+?\\.pm\\)" "/\\1" file-name))
           (find-file file-name)))
        (t
         (hms-perl-create-unit-test
          (hms-perl-make-test-package-name
           (hms-perl-get-package-name-for-current-buffer))))))

(defun hms-perl-create-unit-test (test-package)
  "Create the unit test file for the named package (defaults to current buffer)."
  (interactive (list (read-string "Test Package Name: "
                                  (hms-perl-make-test-package-name 
                                   (hms-perl-get-package-name-for-current-buffer)))))
  (let* ((test-directory "Test")
         (test-file (hms-perl-compute-test-file-for-current-buffer))
         (test-existed (file-exists-p test-file)))
    (message "test-package:%s test-directory:%s test-file:%s" test-package test-directory test-file)
    (if (not (file-exists-p test-directory))
        (progn
          (message "creating directory: %s" test-directory)
          (make-directory test-directory)))
    (find-file test-file)
    (if (not test-existed) (hms-perl-new-module-unit-test test-package))))

(defun hms-perl-base-package-for-test-package (test-package)
  "Remove the ::Test from the test-package name"
  (replace-regexp-in-string "::Test::" "::" test-package))

(defun hms-perl-new-module-unit-test (test-package)
  "Create a new unit test module."
  (interactive
   (list
    (read-string
     "Test Package Name: " 
     (hms-perl-make-test-package-name (hms-perl-get-package-name-for-current-buffer)))))
  (insert "# -*- mode: CPerl -*-\n")
  (insert "package " test-package ";\n")
  (insert "use strict;\n")
  (insert "use warnings;\n")
  (insert "our $CVS_ID = '$Id: emacs-utils.el,v 1.60 2008/03/11 19:58:48 jcrean Exp $';\n")
  (insert "our($VERSION) = ($CVS_ID =~ /\\s(\\d+(?:\\.\\d+)+)/);\n")
  (insert "use Test::Unit::TestCase;\n")
  (insert "use Test::Unit::Assertion::Boolean;\n")
  (insert "use " *hms-perl-standard-base-class* ";\n")
  (insert "use " (hms-perl-base-package-for-test-package test-package) ";\n")
  (insert "use Carp;\n")
  (insert "use base qw( Test::Unit::TestCase HMS::Util::UnitTest );\n")
  (insert "use HMS::Util::UnitTest qw(:All);\n")
  (insert "\n")
  (insert "my $Log = " *hms-perl-standard-base-class* "->get_log_context;\n")
  (insert "\n")
  (insert  *hms-perl-standard-base-class* "->MakeGSS(qw( ));\n")
  (insert "\n")
  (if *hms-perl-include-pod*
      (progn
        (insert "=head1 NAME\n")
        (insert "\n")
        (insert test-package " - Unit test\n")
        (insert "\n")
        (insert "=head1 SYNOPSIS\n")
        (insert "\n")
        (insert "=head1 DESCRIPTION\n")
        (insert "\n")
        (insert "=head1 API REFERENCE\n")
        (insert "\n")
        (insert "=cut\n")
        (insert "\n")
        (insert "# start your test methods here\n")
        (insert "\n")
        (insert "=head2 test_compilation\n")
        (insert "\n")
        (insert "Validate that the module compiles.\n")
        (insert "\n")
        (insert "=cut \n")))
  (insert "\n")
  (insert "sub test_compilation {\n")
  (insert "  my($self) = @_;\n")
  (insert "  $self->assert(1,'Error, compilation failed?');\n")
  (insert "}\n")
  (insert "\n")
  (insert "1;\n")
  (previous-line 2)
  (recenter)
  (message "Don't forget to document the package and the API."))

(defun hms-perl-make-test-package-name (pkg)
  (let* ((parts (split-string pkg "::"))
         (last-part (hms-shift parts)))
    (setq parts (reverse parts))
    (push "Test" parts)
    (push last-part parts)
    (setq parts (reverse parts))
    (let ((test-package (hms-string-join "::" parts)))
      (if (string-match "::Test::Test::" test-package) 
          (replace-match "::Test::" t t test-package)
        test-package))))



(defun hms-perl-package-to-file-name (package)
  (interactive)
  (concat (hms-string-join "/" (split-string package "::")) ".pm"))

(defun hms-perl-new-member (member-name)
  "Add a new member variable (get/set)."
  (interactive (list (read-string "Member Variable: " 
                                  (hms-perl-word-under-cursor))))
  (let ((previous-pos (point)))
    (message "previous-pos: %s" previous-pos)
    (goto-char 1)
    (search-forward "->MakeGSS(")
    (search-forward "qw(")
    (insert " " member-name " ")
    (goto-char (+ previous-pos (length member-name) 1 ))
    (hms-perl-insert-indented "$self->" member-name )))

(defun hms-perl-insert-log-statement (log-level)
  "Inserts a log statment."
  (interactive (list (read-string "Logging Level: (default debug) " nil nil "debug")))
  (hms-perl-insert-indented "$Log->" log-level "(\"\") if $Log->is_" log-level ";")
  (search-backward-regexp "(\"")
  (forward-char 2))

(defun hms-insert-file-name ()
  "Insert the current file's name into the buffer"
  (interactive)
  (insert buffer-file-name))

(defun hms-string-join (delim lat)
  (let ((string (car lat)))
    (loop for item in (cdr lat)
          do
          (setq string (concat string delim item)))
    string))
        
(defun hms-get-pwd-as-list ()
  (split-string buffer-file-name "/"))

(defun hms-get-pwd-as-list-no-lib ()
  (split-string (hms-string-strip-lib-prefix buffer-file-name) "/"))

(defun hms-insert-pwd ()
  "Inserts the pwd based on the current file name."
  (interactive)
  (let ((pwd (hms-string-join "/" (reverse (cdr (reverse 
                                                 (hms-get-pwd-as-list)))))))
    (message "path: %s" pwd)
    (insert pwd)))

(defun hms-compute-package-from-pwd ()
  "Attempts to create a package name from the pwd."
  (interactive ("sPackageName: " ))
  (let* ((pwd (hms-get-pwd-as-list-no-lib))
         (file (hms-shift pwd)))
    (message "pwd=%s" pwd)
    (concat (hms-string-join "::" pwd) "::" (hms-string-strip-file-suffix file))))

(defun hms-file-to-package (file)
  "Converts a file name into a module/class/package name."
  (hms-string-strip-lib-prefix (hms-string-strip-file-suffix file)))

(defun hms-string-strip-file-suffix (name)
  (let ((pos (hms-string-find name "." 0)))
    (cond ((= -1 pos) name)
          (t (substring name 0 pos)))))

(defun hms-string-strip-lib-prefix (name)
  (let ((pos (hms-string-find name *hms-perl-source-path-prefix* 0)))
    (message "pos=%s" pos)
    (cond ((= -1 pos) name)
          (t
           (substring name (+ pos (length *hms-perl-source-path-prefix*)))))))

(defun hms-string-find (buff pat pos)
  (let ((pos 0) (loop t) (max (- (length buff) (length pat))))
    (while (and loop (< pos max))
      (cond ((string= pat (substring buff pos (+ pos (length pat))))
             (setq loop nil))
            (t
             (setq pos (+ 1 pos)))
            ))
    (if (= pos max) -1 pos)))

(defun hms-perl-add-unit-test-for-method ()
  "For the method that the point is currently within, create a test
method in the unit test package."
  (interactive)
  (search-backward-regexp "^sub")
  (forward-word 1)
  (forward-char 1)
  (let ((beg (point))
        (function-name))
    (search-forward-regexp "\\( \\|{\\|\n\\)")
    (backward-char 1)
    (setq function-name (buffer-substring beg (point)))
    (message "function-name: %s" function-name)
    (find-file (hms-perl-compute-test-file-for-current-buffer))
    (hms-perl-insert-method (concat "test_" function-name) "")))

(defun hms-perl-compile-buffer ()
  "Runs 'perl -wc' on the current buffer"
  (interactive)
  (shell-command (concat "perl -I " (getenv "PERL5LIB") " -wc " buffer-file-name)))

(defun hms-perl-plint-buffer ()
  "Runs 'plint' on the current buffer"
  (interactive)
  (shell-command
   (concat "PERL5LIB=\"" (getenv "PERL5LIB") "\" plint --summary " buffer-file-name)))

(defun hms-perl-print-test-runner-cmdline ()
  (interactive)
  (message (hms-perl-make-test-runner-cmdline (hms-perl-get-package-name-for-current-buffer))))

(defun hms-perl-make-test-runner-cmdline (package)
  "Show a simple command line for running the Test::Unit test runner
for the current package."
  (interactive (list (read-string "Package: " (hms-perl-get-package-name-for-current-buffer))))
  (format "PERL5LIB=\"%s\" %s %s 2>&1"
          (getenv "PERL5LIB")
          *hms-perl-test-unit-test-runner-cmd*
          (hms-perl-make-test-package-name 
           package)))

(defun hms-perl-run-unit-test ()
  "Runs unit test for the current module."
  (interactive)
  (let ((test-package (hms-perl-make-test-package-name 
                       (hms-perl-get-package-name-for-current-buffer))))
    (if (string-match "::Test::Test::" test-package) 
        (setq test-package 
              (replace-match "::Test::" t t test-package)))
    (shell-command
     (funcall *hms-perl-unit-test-cmdline-func*
              test-package))))

(defun hms-perl-make-test-suite-runner-cmdline ()
  (interactive)
  (format "PERL5LIB=\"%s\" %s 2>&1"
          (getenv "PERL5LIB")
          *hms-perl-test-suite-runner-cmd*))

(defun hms-perl-run-test-suite ()
  "Runs the full test suite"
  (interactive)
  (shell-command
   (funcall *hms-perl-test-suite-cmdline-func*)))
   

;;
;; Functions for interacting with Log4perl
;;

(defun hms-perl-package-to-logger-entry (pkg)
  (concat "log4perl.logger." (hms-string-join "." (split-string pkg "::") )))

(defun hms-perl-open-log4perl-config ()
  "Open the Log4perl configuration file in a buffer."
  (interactive)
  (find-file *hms-perl-log4perl-config-file*))

(defun hms-perl-set-new-log4perl-config (config-file)
  "Set a new log4perl configuration file (sets environment variable)."
  (interactive (list (read-string "Set Log4perl Config: " 
                                  *hms-perl-log4perl-config-file* nil *hms-perl-log4perl-config-file*)))
  (setenv *hms-perl-log4perl-config-env-var* config-file))

(defun hms-perl-adjust-logging-for-package (level)
  "Adjust the logging level for the current package. Opens the file specified 
by *hms-perl-log4perl-config-env-var* and updates the context line, 
inserting one if not found."
  (interactive (list (read-string "Set Log Level: (default DEBUG, * to remove) " nil nil "DEBUG")))
  (let ((logger-context (hms-perl-package-to-logger-entry 
                         (hms-perl-get-package-name-for-current-buffer))))
    (find-file *hms-perl-log4perl-config-file*)
    (goto-char 1)
    (if (search-forward logger-context nil t)
        (progn
          (beginning-of-line)
          (kill-line))
      (progn
        (end-of-buffer)
        (insert "\n")
        (beginning-of-line)))
    (message "level=%s" level)
    (if (equal "*" level)
        (progn (kill-line))
      (progn (insert logger-context "=" level "\n")
             (backward-char 1)))
    (save-buffer)
    (kill-buffer (current-buffer))))

(defun hms-perl-load-agent-env ()
  (interactive)
  (find-file (expand-file-name "~/.agent.env"))
  (goto-char 1)
  (search-forward "=")
  (let ((beg (point)))
    (search-forward ";")
    (backward-char 1)
    (setenv "SSH_AUTH_SOCK" (buffer-substring beg (point))))
  (next-line 1)
  (beginning-of-line)
  (search-forward "=")
  (let ((beg (point)))
    (search-forward ";")
    (backward-char 1)
    (setenv "SSH_AGENT_PID" (buffer-substring beg (point))))
  (kill-buffer (current-buffer)))


(defun hms-perl-insert-exception (type)
  "Insert a basic throw statement."
  (interactive (list (read-string "Exception Type ([HMS::]Exception) " nil nil 
                                  "Exception")))
  (if (not (string-match "^HMS::" type))
      (setq type (concat "HMS::" type)))
  (hms-perl-insert-indented "HMS::Base->throw(" type "->new(join('',\"\")));")
  (backward-char 5)
  (cperl-indent-command))

(defun hms-perl-insert-indented-string (text)
  (let ((start (point)))
    (insert text)
    (cperl-indent-region start (point)))
  (end-of-line))

(defun hms-perl-insert-indented (&rest args)
  (mapcar #'hms-perl-insert-indented-string args))

(defun hms-perl-insert-try-catch ()
  "Insert a basic try/catch block"
  (interactive)
  (let ((pos))
    (hms-perl-insert-indented "eval {\n")
    (setq pos (point))
    (hms-perl-insert-indented " \n")
    (hms-perl-insert-indented "};\n")
    (hms-perl-insert-indented "\n")
    (hms-perl-insert-indented "if ($@) {\n")
    (hms-perl-insert-indented "}\n")
    (hms-perl-insert-indented "\n")
    (goto-char pos)
    (cperl-indent-command)
    (end-of-line)))

(defun hms-perl-perldoc-buffer ()
  "Execute pod2text on the current buffer, opening a new persistient
buffer for the output.  Subsequent invocations will re-use the same buffer."
  (interactive)
  (let ((module-buffer-name buffer-file-name))
    (switch-to-buffer (get-buffer-create "perldoc"))
    (beginning-of-buffer)
    (let ((beg (point)))
      (end-of-buffer)
      (kill-region beg (point)))
    (insert (shell-command-to-string (concat "PERL5LIB=\"" (getenv "PERL5LIB") "\" pod2text " module-buffer-name)))
    (beginning-of-buffer)
    ))

(defun hms-perl-string-remove-regex (pat string)
  (if (string-match pat string)
      (let ((start (match-beginning 0)) (end (match-end 0)))
        (concat (substring string 0 start) (substring string end))) string))

(defun hms-perl-convert-print-to-log ()
  "Convert the print statement to a Log->* statement. The print
statement that the point is currently within will be converted. This
function operates by first searching left for 'print' to find the
beginning of the statement, then to the right for a line ending with a
semi-colon. This function does not handle printf, only print."
  (interactive)
  (let ((start) (end) (fh-name) (statement) (log-type "debug"))
    (search-backward-regexp "\\bprint[^f]")
    (setq start (point))
    (search-forward-regexp ";\\s-*")
    (setq end (point))
    (setq statement (buffer-substring start end))
    (kill-region start end)
    (if (string-match "^print\\s-*(?\\s-*\\(STDERR\\|STDOUT\\)\\s-+" statement)
        (progn
          (setq fh-name (match-string 1 statement))
          (cond ((string-equal "STDOUT" fh-name)
                 (setq log-type "debug"))
                ((string-equal "STDERR" fh-name)
                 (setq log-type "warn")))))
    (setq statement (hms-perl-string-remove-regex "^print\\s-*(?\\s-*" statement))
    (setq statement (hms-perl-string-remove-regex "^\\(STDERR\\|STDOUT\\)\\s-*" statement))
    (setq statement (hms-perl-string-remove-regex "^\\s-+" statement))
    (setq statement (hms-perl-string-remove-regex "\\s-*;\\s-*" statement))
    (hms-perl-insert-indented "$Log->" log-type "(" statement ") if $Log->is_" log-type ";")))


(defun hms-perl-justify-line-at-token (substr column)
  "Internaly justfy a line's contents based on a known token/seperator."
  (interactive (list (read-string "Token (=>): " nil nil "=>")
                     (read-string "Column (*): " nil nil "*")))
  (cond ((string-equal "*" column) ;; defautlt to the current column
         (setq column (current-column)))
        (t
         (setq column (string-to-number column))))
  (let ((left-part) (right-part) (start) (end) (pos))
    (beginning-of-line)
    (setq start (point))
    (end-of-line)
    (setq end (point))
    (setq left-part (buffer-substring start end))
    (setq pos (hms-string-find left-part substr 0))
    (setq right-part (substring left-part (+ pos (length substr))))
    (setq left-part (substring left-part 0 pos))
    (kill-region start end)
    (insert (concat (hms-perl-string-rpad left-part column " ") substr 
                    right-part))))


(defun hms-perl-string-rpad (str len chr)
  "Simple right pad of str out to len by apending chr on repeatedly."
  (while (> len (length str))
    (setq str (concat str chr)))
  str)

(defun hms-perl-string-lpad (str len chr)
  "Simple left pad of str out to len by apending chr on repeatedly."
  (while (> len (length str))
    (setq str (concat chr str)))
  str)


(defun hms-perl-concat-env (var delim value)
  "Push an additional element into your environment with specified delimiter"
  (interactive "sEnvironment Variable: \nsDelimiter: \nsValue: ")
  (if (getenv var )
      (setenv var (concat (getenv var) delim value))
    (setenv var value))
  t)

(defun hms-perl-push-inc ( path )
  "Push additional element onto PERL5LIB"
  (interactive "DPath: ")
  (hms-perl-concat-env "PERL5LIB" ":" (expand-file-name path)))

(defun hms-perl-insert-block-conditional-construct (name)
  "For inserting a block conditional syntax construct like an if, a
while, or an unless."
  (hms-perl-insert-indented
   (hms-string-join "\n"
                    (list
                     (concat name " () {")
                     "}")))
  (search-backward-regexp ")"))

(defun hms-perl-insert-if ()
  "Insert an if statement..."
  (interactive)
  (hms-perl-insert-block-conditional-construct "if"))

(defun hms-perl-insert-unless ()
  "Insert an unless statement..."
  (interactive)
  (hms-perl-insert-block-conditional-construct "unless"))

(defun hms-perl-insert-while ()
  "Insert an while statement..."
  (interactive)
  (hms-perl-insert-block-conditional-construct "while"))

(defun hms-perl-add-else-clause ()
  "Add an else clause to the current block."
  (interactive)
  (search-forward-regexp "}")
  (hms-perl-insert-indented " else {")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "\n")
  (hms-perl-insert-indented "}\n")
  (previous-line 2)
  (cperl-indent-command))

(defun hms-perl-insert-foreach (variable list)
  "Insert a foreach statement..."
  (interactive "sVariable: \nsList: ")
  (hms-perl-insert-indented
   (hms-string-join "\n"
                    (list
                     (concat "foreach my " variable " (" list ") {")
                     ""
                     "}")))
  (previous-line 1)
  (cperl-indent-command))

(defun hms-perl-copy-and-comment-region ()
  (interactive)
  (let ((start (point)) (end (mark)))
    (save-excursion
      (kill-region start end)
      (hms-perl-insert-indented (concat *hms-perl-reworked-block-comment-start* 
                                        "\n"))
      (setq start (point))
      (yank)
      (setq end (point))
      (comment-region start end)
      (hms-perl-insert-indented (concat *hms-perl-reworked-block-comment-mid* 
                                        "\n"))
      (yank)
      (hms-perl-insert-indented (concat *hms-perl-reworked-block-comment-end* 
                                        "\n")))))

(defun hms-perl-get-package-name-for-current-buffer ()
  "Starting at the beginning of the buffer, look for the
first instance of a package declaration at the beginning of a line."
  (interactive)
  (let ((package-name))
    (save-excursion
      (beginning-of-buffer)
      (search-forward-regexp "^package ")
      (let ((start (point)))
        (search-forward-regexp ";")
        (setq package-name (buffer-substring start (- (point) 1)))))
    (or package-name (hms-compute-package-from-pwd))))


(defun hms-perl-find-undeclared-perl-variables-in-current-selection ()
  (interactive)
  (hms-perl-find-undeclared-perl-variables-in-string (buffer-substring (point) 
                                                                       (mark))))

(defun hms-perl-refactor-selection-as-local-variable (varname)
  "Takes current selection and refactors as var name"
  (interactive (list (read-string "Variable name: ")))
  (save-excursion
    (let ((rval (buffer-substring (point) (mark))))
      (kill-region (point) (mark))
      (insert varname)
      (beginning-of-line)
      (hms-perl-insert-indented
       (concat "my " varname " = " rval ";\n")))))

(defun hms-perl-refactor-selection-as-method (method-name argument-list)
  "Using the current point and mark as a selection, refactor the code block as
a new method in the current buffer."
  (interactive (list (read-string "Method Name: " "refactoredBlock" )
                     (read-string "Arguments: "
                                  (hms-string-join "," 
                                                   (hms-perl-find-undeclared-perl-variables-in-current-selection)))))

  (let ((code-block (buffer-substring (point) (mark))))
    (kill-region (point) (mark))
    (if argument-list
        (hms-perl-insert-indented
         (concat "$self->" method-name "("
                 argument-list
                 ");\n"))
      (hms-perl-insert-indented (concat "$self->" method-name ";\n")))
    (hms-perl-insert-method method-name argument-list)
    (search-forward-regexp "->throw")
    (beginning-of-line)
    (kill-line)
    (hms-perl-insert-indented code-block)))

(defun hms-perl-find-undeclared-perl-variables-in-string (code-block)
  "Attempt to find any undeclared variables in the given string of code."
  (interactive "sSource Code: ")
  (hms-perl-unique-list
   (hms-perl-find-all-string-match *hms-perl-regex-perlvariables* code-block)))

(defun hms-perl-unique-list (list)
  (let ((new-list))
    (loop for item in list
          do
          (if (not (member item new-list))
              (push item new-list)))
    (reverse new-list)))

(defun hms-perl-find-all-string-match (regex string)
  (let ((items '()) (old-case-fold-search case-fold-search))
    (setq case-fold-search nil)
    (do ((pos
          (string-match regex string)
          (string-match regex string (+ 1 (match-end 0)))))
        ((not pos) items)
      (if pos
          (progn
            (push (substring string (match-beginning 0) (match-end 0)) items))))
    (setq case-fold-search old-case-fold-search)
    items))

(defun hms-perl-get-prefix-of-package (package)
  (let ((parts (split-string package "::")))
    (hms-shift parts)
    (hms-string-join "::" parts)))

(defun hms-perl-refactor-push-method-to-parent (parent-class)
  (interactive (list
                (read-string
                 "Parent Class: "
                 (hms-perl-get-prefix-of-package 
                  (hms-perl-get-package-name-for-current-buffer)))))
  (let ((sub-name (hms-perl-get-name-of-current-sub))
        (start (hms-perl-find-start-of-current-sub-and-haxygen))
        (end (hms-perl-find-end-of-current-sub))
        (code-block)
        (parent-file-name (hms-perl-package-to-file-name parent-class)))
    (setq code-block (buffer-substring start end))
    (kill-region start end)
    (find-file parent-file-name)
    (hms-perl-position-pointer-at-method-insertion-point)
    (hms-perl-insert-indented (concat "\n" code-block "\n"))))

(defun hms-perl-find-start-of-current-sub-and-haxygen ()
  (interactive)
  (let ((position)
        (sub-name (hms-perl-get-name-of-current-sub)))
    (save-excursion
      (if
          (not
           (or
            (search-backward-regexp "#//")
            (search-backward-regexp (concat "sub " sub-name))))
          (error "Error, unable to find begining of current method"))
      (setq position (point)))
    position))

(defun hms-perl-find-end-of-current-sub ()
  (interactive)
  (let ((position)
        (sub-name (hms-perl-get-name-of-current-sub)))
    (save-excursion
      (if
          (not
           (search-backward-regexp (concat "sub " sub-name)))
          (error "Error, unable to find begining beginning of current method (when 
trying to find end)."))
      (search-forward-regexp "{")
      (backward-char 1)
      (match-paren nil)
      (forward-char 1)
      (setq position (point)))
    position))

(defun hms-perl-refactor-eliminate-local (variable-name expression)
  "Eliminate a local variable in leiu of using the literal expression (simple 
string replacement)."
  (interactive (list
                (read-string "Variable Name: " "")
                (read-string "Value: " "")))
  (let* ((start (hms-perl-find-start-of-current-sub-and-haxygen))
         (end (hms-perl-find-end-of-current-sub))
         (code-block (buffer-substring start end)))
    (setq code-block (replace-regexp-in-string variable-name expression 
                                               code-block))
    (kill-region start end)
    (hms-perl-insert-indented code-block)))

(defun hms-perl-refactor-introduce-local (variable-name expression)
  "Introduce a local variable in leiu of a literal expression (simple string 
replacement)."
  (interactive (list
                (read-string "Variable Name: " "")
                (read-string "Value: " "")))
  (beginning-of-line)
  (hms-perl-insert-indented "\n")
  (backward-char 1)
  (hms-perl-insert-indented "--placeholder-replacement-string--;\n")
  (let* ((start (hms-perl-find-start-of-current-sub-and-haxygen))
         (end (hms-perl-find-end-of-current-sub))
         (code-block (buffer-substring start end)))
    (setq code-block (replace-regexp-in-string expression variable-name 
                                               code-block))
    (setq code-block
          (replace-regexp-in-string
           "--placeholder-replacement-string--"
           (concat "my " variable-name " = \"" expression "\"")
           code-block))

    (kill-region start end)
    (hms-perl-insert-indented code-block)))

(defun hms-perl-package-under-cursor ()
  "Returns the 'package' at the point."
  (interactive)
  (save-excursion
    (search-backward-regexp "[^a-zA-Z0-9:]")
    (forward-char 1)
    (let ((beg (point)))
      (search-forward-regexp "[^a-zA-Z0-9:]")
      (backward-char 1)
      (buffer-substring beg (point)))))

(defun hms-perl-get-entire-buffer-as-string ()
  "Return the entire buffer as a single string."
  (interactive)
  (let ((data nil))
    (save-excursion
      (end-of-buffer)
      (setq data (buffer-substring 1 (point))))
    data))

(defun hms-perl-find-all-packages-in-buffer ()
  (interactive)
  "Using a regex based search find all the package names in the current 
buffer."
  (remove (hms-perl-get-package-name-for-current-buffer)
          (hms-perl-unique-list
           (hms-perl-find-all-string-match
            *hms-perl-regex-perlpackages*
            (hms-perl-get-entire-buffer-as-string)))))

(defun hms-perl-string-is-already-present-in-buffer (string)
  "Searches for the string starting from the beginning of the buffer.
Returns true if found, false if not."
  (save-excursion
    (beginning-of-buffer)
    (search-forward string (point-max) t)))

(defun hms-perl-fix-use-modules ()
  "Insert a 'use module' statement at the top of the
package (two lines after the __PACKAGE__->MakeGSS') for each
module (perl package name) mentioned in the source code.

Additionaly prompt the user for removal of any modules that are used
but don't look to be actualy mentioned in the buffer."
  (interactive)
  ;; FIXME method is too long
  ;; FIXME it catche's too much stuff (OBJECT::TYPE, and misses other things like Fcntl)
  (let ((modules (hms-perl-find-all-packages-in-buffer))
        (inserted '()))
    (save-excursion
      (beginning-of-buffer)
      (search-forward-regexp "->MakeGSS(")
      (search-forward-regexp "));")
      (forward-line 1)
      (loop for module in modules
            do
            (progn
              (let ((line (concat "\nuse " module)))
                (when (not (hms-perl-string-is-already-present-in-buffer line))
                  (end-of-line)
                  (hms-perl-insert-indented (concat line ";"))
                  (push module inserted))))))
    (message "Added use statements for %s" inserted))
  ;; second part is to look for all use statements and suggest that
  ;; any that don't look relevant be removed...
  (let ((modules (hms-perl-find-all-unused-modules-in-buffer))
        (answer))
    (if modules
        (setq answer (read-string (format "Remove modules %s [y/n]? " modules)))
      (message "All modules look to be used, nothing to remove."))
    (when (and answer (string-equal "y" (substring answer 0 1)))
      (message "Removing %s" modules)
      (save-excursion
        (loop for module in modules
              do
              (progn
                (beginning-of-buffer)
                (search-forward (concat "use " module ";"))
                (beginning-of-line)
                (kill-line)
                (delete-char 1)))))))

(defun hms-perl-insert-identity-constant (constantName)
  "Insert value into HMS::Base->MakeIdentityConstants(qw(...)) list (create this 
line if it doesn't exist)"
  (interactive (list (read-string "Constant: " (hms-perl-word-under-cursor))))
  (save-excursion
        (if (not (string= constantName (hms-perl-word-under-cursor)))
            (insert constantName "()"))
        (beginning-of-buffer)
        (if 
            (not (search-forward "->MakeIdentityConstants(qw(" (point-max) t))
            (progn
              (search-forward "MakeGSS")
              (forward-line)
              (insert (concat "\nHMS::Base->MakeIdentityConstants(qw());\n"))))
        (beginning-of-buffer)
        (search-forward "->MakeIdentityConstants(qw(")
        (let ((start (point))
              (const-list nil))
          (search-forward ")")
          (backward-char)
          (setf const-list (buffer-substring start (point)))
          (kill-region start (point))
          (insert (hms-string-join " " (hms-perl-unique-list (split-string (concat const-list " " constantName) " ")))))))

(defun hms-perl-find-all-unused-modules-in-buffer ()
  "Look for all use statements (of the form `use Module::Name;'), and
return the list of any that are not actualy mentioned in the buffer."
  (interactive)
  (let ((whole-buffer (hms-perl-get-entire-buffer-as-string))
        (lines '())
        (modules))
    (setq lines (hms-perl-find-all-string-match
                 (concat "use " *hms-perl-regex-perlpackages* ";")
                 whole-buffer))
    (loop for line in lines
          do
          (let ((module-name (substring line 4)))
            (setq module-name (replace-regexp-in-string ";" "" module-name))
            (if (> 2 (length (hms-perl-find-all-string-match (concat module-name 
                                                                     "[^_:a-zA-Z]") whole-buffer)))
                (push module-name modules))))
    (hms-perl-unique-list modules)))

(defun hms-perl-get-all-constant-decls-from-buffer ()
  "Seraches for constant declarations in the current buffer and
returns the constant names as a list."
  (interactive)
  (save-excursion
    (let ((whole-buffer (hms-perl-get-entire-buffer-as-string))
          (found-constants '())
          (constants '()))
      (setq found-constants (hms-perl-find-all-string-match
                             *hms-perl-regex-constant-decls*
                             whole-buffer))
      (loop for const in found-constants
            do
            (progn
              (setq const (substring const (length "use constant ")))
              (setq const (substring const 0 (- (length const) 3)))
              (setq const (replace-regexp-in-string " \+$" "" const))
              (push const constants)))
      constants)))

(defun hms-perl-insert-all-constants-as-list ()
  "Inserts the list of all constants declared in the current buffer at
the point."
  (interactive)
  (insert (hms-string-join " " 
                           (hms-perl-get-all-constant-decls-from-buffer))))

(defun hms-perl-export-all-constants ()
  "Looks for '%EXPORT_TAGS', then 'Constants => [qw(' and will fill
the quotes word list with all constants declared in the buffer."
  (interactive)
  (save-excursion
    (beginning-of-buffer)
    (search-forward "%EXPORT_TAGS")
    (search-forward "Constants => [qw( ")
    (let ((start (point)))
      (search-forward " )]")
      (kill-region start (- (point) 3))
      (search-backward " )")
      (hms-perl-insert-all-constants-as-list))))

(defun hms-perl-find-next-arg-list ()
  "Find the next thing that looks like an arg list from the point forward."
  (interactive)
  (condition-case nil
      (save-excursion
        (let ((start))
          (search-forward "(")
          (setq start (point))
          (search-forward ")")
          (backward-char 1)
          (buffer-substring start (point))))
    (error nil)))

(defun hms-perl-insert-join ()
  (interactive)
  (hms-perl-insert-indented "join(\",\",)")
  (backward-char 1))

(defun hms-perl-string-left-trim (string)
  (while (string-match "^ " string)
    (setq string (substring string 1)))
  string)

(defun hms-perl-string-right-trim (string)
  (while (string-match " $" string)
    (setq string (substring string 0 -1)))
  string)

(defun hms-perl-string-trim (string)
  (hms-perl-string-right-trim
   (hms-perl-string-left-trim string)))

(defun hms-perl-open-new-line-above ()
  "Kinda like C-o, but jumps to beginning of line."
  (interactive)
  (beginning-of-line)
  (open-line 1)
  (cperl-indent-command))

(defun hms-perl-rename-current-sub (old-name new-name)
  (interactive (list
                (read-string "Sub: " (hms-perl-get-name-of-current-sub))
                (read-string "New Name: " (hms-perl-get-name-of-current-sub))))
  (save-excursion
        (beginning-of-buffer)
        (query-replace old-name new-name)))


(defun hms-perl-insert-method-and-test-method (method-name arg-list)
  (interactive (list (read-string "Method Name: " (hms-perl-word-under-cursor))
                     (read-string "Arguments in addition to $self: " 
                                  (hms-perl-find-next-arg-list))))
  (hms-perl-insert-method method-name arg-list)
  (search-forward "my(")
  (hms-perl-add-unit-test-for-method))

(defun hms-perl-buffer-between-regexes (start-regex end-regex)
  "Given two bounding regexes, attempts to extract the region inbetween them."
  (interactive)
  (let ((pkg ""))
    (save-excursion
      (search-backward-regexp start-regex)
      (goto-char (match-end 0))
      (let ((start (point)))
        (search-forward-regexp end-regex)
        (backward-char (length (match-string 0)))
        (setf pkg (buffer-substring start (point)))))
    pkg))

(defun hms-perl-package-at-point ()
  "Find the package name under (close to) the point."
  (interactive)
  (hms-perl-buffer-between-regexes "[^a-zA-Z_0-9:]" "[^a-zA-Z_0-9:]"))

(defun hms-perl-get-log-line-pattern-package-line ()
  (let ((package "")
        (line-no -1)
        (started-at (point)))
    (search-backward-regexp "[^a-zA-Z_0-9:]")
    (forward-char (length (match-string 0)))
    (let ((start (point)))
      (goto-char started-at)
      (search-forward-regexp "(")
      (backward-char (length (match-string 0)))
      (setf package (buffer-substring start (point))))
    (forward-char 1)
    (let ((start (point)))
      (search-forward-regexp ")")
      (backward-char 1)
      (setf line-no (buffer-substring start (point))))
  (goto-char started-at)
  (list package line-no)))

(defun hms-perl-jump-to-file-and-line (file line)
  (interactive "File: \nnLine: ")
  (find-file file)
  (switch-to-buffer (find-buffer-visiting file))
  (goto-line line))

(defun hms-perl-jump-to-class-and-sub (class sub)
  (interactive "sClass: \nsMethod: ")
  (find-file (hms-perl-find-package-file class))
  (beginning-of-buffer)
  (search-forward-regexp (format "sub %s" sub)))

(defun hms-perl-jump-to-class-and-line ()
  (interactive)
  (let* ((pkg-line (hms-perl-get-log-line-pattern-package-line))
         (package (first pkg-line))
         (line-no (second pkg-line))
         (file-name
          (cond ((hms-perl-find-package-file package)
                 (hms-perl-find-package-file package))
                (t
                 (hms-perl-find-package-file 
                  (replace-regexp-in-string "::[^:]+$" "" package))))))
    (hms-perl-jump-to-file-and-line file-name (read-from-string line-no))))

(defun hms-perl-find-package-file (package)
  "Search the PERL5LIB for the file for the given package name."
  (interactive
   (list
    (read-string "Package: " (hms-perl-package-at-point))))
  (let ((paths (split-string (getenv "PERL5LIB") ":"))
        (sub-path (concat
                   (replace-regexp-in-string
                    "::" "/" package)
                   ".pm")))
    (loop for path in paths
          do
          (let ((candidate (concat path "/" sub-path)))
            (when (file-exists-p candidate)
              (return candidate))))))

(defun hms-perl-open-buffer-for-package (package)
  "Open the file for the package under the point (pushing onto the tags stack)"
  (interactive
   (list
    (read-string "Package: " (condition-case nil
                                 (hms-perl-package-at-point)
                               (error "")))))
  (let ((file-name (hms-perl-find-package-file package)))
    (push (buffer-file-name) *hms-perl-tags-file-stack*)
    (find-file file-name)))

(defun hms-perl-pop-tags-stack ()
  "Return to the file jumped _from_ with hms-perl-open-buffer-for-package."
  (interactive)
  (find-file (pop *hms-perl-tags-file-stack*)))

(defun hms-perl-get-method-at-point ()
  "Attempts to extract just a method name from around the point."
  (interactive)
  (hms-perl-buffer-between-regexes "[^a-zA-Z_0-9]" "[^a-zA-Z_0-9]"))

(defun hms-perl-get-method-call-at-point ()
  "Attempts to extract a method call from around the point."
  (interactive)
  (let ((meth "")
        (started-at (point)))
    (save-excursion
      (search-backward-regexp "[ ;]")
      (forward-char (length (match-string 0)))
      (let ((start (point)))
        (goto-char started-at)
        (search-forward-regexp "[^a-zA-Z_0-9]")
        (backward-char (length (match-string 0)))
        (setf meth (buffer-substring start (point)))))
    meth))

;; (defun hms-perl-get-base-classes-for-buffer ()
;;   (interactive)
;;   ;; support either 'use base' or '@ISA'
;;   (save-excursion
;;     (beginning-of-buffer)
;;     (search-forward-regexp

(defun hms-perl-search-inheritence-for-method-declaration (method-name)
  (interactive)
  (message "(inheritence) search for: %s" method-name))
         
(defun hms-perl-search-for-method-declaration (method-name)
  (interactive)
  ;; first search the imported modules in the current buffer...
  (let* ((command (format "fgrep -l -R 'sub %s' $(echo $PERL5LIB | tr : \\\\n) 2>/dev/null" method-name))
         (output (shell-command-to-string command))
         (lines (split-string output "\n")))
    (cond ((= 1 (length lines))
           (find-file (first lines))
           (search-forward-regexp (format "sub %s" method-name)))
          (t
           (message "(%s) found more than 1: %s" command lines)))))

(defun hms-perl-find-next-error ()
  (interactive)
  (set-buffer *hms-perl-test-buffer-name*)
  (condition-case nil
      (progn
        (search-forward-regexp "\\(/.+\\)\\(:\\| line \\)\\([0-9]+\\)")
        (hms-perl-jump-to-file-and-line (match-string 1) (read-from-string (match-string 3))))
    (error (beginning-of-buffer))))

(defmacro hms-perl-extract-buffer-between (start-cmds end-cmds)
  `(let ((result ""))
     (save-excursion
       ,@start-cmds
       (let ((started-at (point)))
         ,@end-cmds
         (setf result (buffer-substring started-at (point)))))
     result))

(defun hms-perl-get-stack-trace-from-buffer ()
  (interactive)
  (hms-perl-extract-buffer-between
   ((search-forward-regexp "Stack Trace:")
    (forward-line 1)
    (beginning-of-line))
   ((search-forward-regexp "Process Id: ")
    (previous-line 2)
    (end-of-line))))

(defun hms-perl-string-starts-with (str substr)
  (= 0 (hms-string-find str substr 0)))

(defun hms-perl-string-contains (str substr)
  (not (= -1 (hms-string-find str substr 0))))

(defun hms-perl-follow-stack-trace ()
  (interactive)
  (let ((stack-lines 
         (remove-if (lambda (line)
                      (string-match "(eval [^)]+)([0-9]+)" line))
                    (split-string (hms-perl-get-stack-trace-from-buffer) "\n")))
        (initial-buffer (current-buffer))
        (go-back-to-initial-buffer t)
        (stop nil)
        (pos 0))
    (while (and (not stop) (not (null stack-lines)))
      (let ((line (nth pos stack-lines))
            (class "")
            (file "")
            (line-no ""))
        (message "line: %s" line)
        (string-match " \\([a-zA-Z_0-9:]+\\)->\\(.+\\)(\\([0-9]+\\))" line)
        (setf class (match-string 1 line))
        (setf file (match-string 2 line))
        (setf line-no (match-string 3 line))
        (message "class:%s file:%s line-no:%s" class file line-no)
        (hms-perl-jump-to-file-and-line file (read-from-string line-no))
        (let ((response (read-char "(x) exit, (q) stop here, (n) next, (p) prev:")))
          (cond ((or (char-equal ?q response)
                     (char-equal ?Q response)
                     (char-equal ?\r response)
                     (char-equal ?\n response))
                 (setf go-back-to-initial-buffer nil)
                 (setf stop t))
                ((or (char-equal ?x response)
                     (char-equal ?X response))
                 (setf stop t))
                ((or (char-equal ?n response)
                     (char-equal ?N response))
                 (if (= pos (- (length stack-lines) 1))
                     (progn (message "at bottom of stack") (sleep-for 0 500))
                   (setf pos (+ 1 pos))))
                ((or (char-equal ?t response)
                     (char-equal ?T response))
                 (setf pos 0))
                ((or (char-equal ?p response)
                     (char-equal ?P response))
                 (if (not (= pos 0))
                     (setf pos (- pos 1))
                   (progn (message "at top of stack") (sleep-for 0 500))))
                ;; anything else quits
                (t
                 t)))))
  (if go-back-to-initial-buffer
      (switch-to-buffer initial-buffer))))

(defun hms-perl-follow-unit-test-stack-trace ()
  (interactive)
  (set-buffer *hms-perl-test-buffer-name*)
  (beginning-of-buffer)
  (hms-perl-follow-stack-trace))


(defun hms-perl-quote-literal (literal)
  "Quote the value if necessary.

  foo => 'foo'
  \"foo\" => \"foo\"
  'foo' => 'foo'
  3 => 3
  3abc => '3abc'
"
  (cond ((string-match "^'.*'$" literal)
         literal)
        ((string-match "^\".*\"$" literal)
         literal)
        ((string-match "^[0-9]+\.?[0-9]*$" literal)
         literal)
        (t
         (concat "'" literal "'"))))

;; (hms-perl-quote-literal "3")
;; (hms-perl-quote-literal "3.14")
;; (hms-perl-quote-literal "3.14abc")
;; (hms-perl-quote-literal "foo")

(defun hms-perl-insert-constant (constant-name constant-value)
  (interactive
   (list
    (read-string "Constant: " "")
    (read-string "Value: " (hms-perl-word-under-cursor))))
  (save-excursion
    (beginning-of-buffer)
    (search-forward-regexp "->MakeGSS")
    (search-forward-regexp ");")
    (forward-line)
    (insert "\nuse constant " constant-name " => " (hms-perl-quote-literal constant-value) ";\n"))
  (insert constant-name "()"))

  

